/*
 * pgpSHADouble.c - Double width version of the NIST SHA-1 hash
 *
 * Used for generalizations of DSA signatures up to 4K bits
 *
 * We use a variation of the modified Benes (double butterfly) construction
 * described in "Foiling Birthday Attacks in Length-Doubling Transformations",
 * William Aiello and Ramarathnam Venkatesan, Eurocrypt 96.
 *
 * The idea is that we perform the following hashes:
 *
 *              Half            Half
 *              Input		   Input
 *                |  \        /  |
 *                |  SHA1  SHA2  |
 *              SHA0    \   /   SHA3
 *                |       X      |
 *                |     /   \    |
 *                |   /       \  |
 *               XOR   	   	    XOR
 *                |  \        /  |
 *                |  SHA5   SHA6 |
 *              SHA4    \   /   SHA7
 *                |       X      |
 *                |     /   \    |
 *                |   /       \  |
 *               XOR   	   	    XOR
 *                |              |
 *              Left Half     Right Half
 *
 *
 * The numbers on the hashes above represent differently keyed SHA-1 hashes.
 * We key each one differently by prepending different numbers of bytes of
 * zeros.
 *
 * $Id: pgpSHADouble.c,v 1.5 1997/10/14 01:48:24 heller Exp $
 */

#include <string.h>
#include <stdio.h>

#include "usuals.h"
#include "sha1.h"
#include "shadbl.h"

/*
 * Size of our odd-even buffers.  Doesn't really matter how big it is.
 * Choose 64 to have a match with the SHA input buffer size, but that is
 * not important.
 */
#define BUFSIZE		64

/* Initialize the SHA values */

void shaDoubleInit(SHADoubleContext *ctx)
{
	byte zeros[3];

	SHA1Init (&ctx->sha0);
	SHA1Init (&ctx->sha1);
	SHA1Init (&ctx->sha2);
	SHA1Init (&ctx->sha3);

	/* Preload hashes with n bytes of 0's to key them differently */
	memset( &zeros, 0, sizeof(zeros) );
	SHA1Update( &ctx->sha1, zeros, 1 );
	SHA1Update( &ctx->sha2, zeros, 2 );
	SHA1Update( &ctx->sha3, zeros, 3 );

	ctx->odd = FALSE;		/* Byte 0 will count as even */
}


/* Update SHA for a block of data. */

void shaDoubleUpdate(SHADoubleContext *ctx, void const *bufParam, word32 len)
{
	byte evenbuf[BUFSIZE], *evenp;		/* Even bytes of buf */
	byte oddbuf[BUFSIZE], *oddp;			/* Odd bytes of buf */
	const byte *buf = (const byte *) bufParam;
	
	/* Divide the bytes of the input into two buffers, one to hold the
	 * even bytes and one to hold the odd bytes.  (The first byte is
	 * counted as even.)  We pass the even bytes through SHA's 0 and 1,
	 * and the odd bytes through 2 and 3.
	 */
	oddp = oddbuf;
	evenp = evenbuf;

#if 0
	/* Naive algorithm for splitting even/odd bytes */

	newodd = ctx->odd ^ (len & 1);
	if( ctx->odd && len != 0 ) {
		*oddp++ = *buf++;
		len--;
	}
	ctx->odd = newodd;

	while ( len != 0 ) {

		while( oddp < oddbuf + BUFSIZE ) {
			if (len-- == 0)
				break;
			*evenp++ = *buf++;
			if (len-- == 0)
				break;
			*oddp++ = *buf++;
		}
		SHA1Update( &ctx->sha0, evenbuf, evenp-evenbuf );
		SHA1Update( &ctx->sha1, evenbuf, evenp-evenbuf );
		SHA1Update( &ctx->sha2, oddbuf, oddp-oddbuf );
		SHA1Update( &ctx->sha3, oddbuf, oddp-oddbuf );
	
	}

#else
	/* If we were left at an odd point last time, handle first byte */
	if( ctx->odd && len != 0) {
		*oddp++ = *buf++;
		/* We intentionally don't decrement len here, handled below */
	}
	ctx->odd ^= (len & 1);

	/* Deal with blocks of data which fill our odd/even buffers */
	while( len >= 2*BUFSIZE ) {
		while (oddp < oddbuf + BUFSIZE ) {
			*evenp++ = *buf++;
			*oddp++ = *buf++;
		}
		//assert( oddp == oddbuf + BUFSIZE );
		//assert( evenp == evenbuf + BUFSIZE ||
		//		   evenp == evenbuf + BUFSIZE - 1 );
		SHA1Update( &ctx->sha0, evenbuf, evenp-evenbuf );
		SHA1Update( &ctx->sha1, evenbuf, evenp-evenbuf );
		SHA1Update( &ctx->sha2, oddbuf, BUFSIZE );
		SHA1Update( &ctx->sha3, oddbuf, BUFSIZE );
		len -= BUFSIZE + (evenp-evenbuf);
		oddp = oddbuf;
		evenp = evenbuf;
	}
		
	/* Deal with leftover bytes of input which are less than our block */
	if (oddp != oddbuf)
		len -= 1;

	while (oddp < oddbuf + BUFSIZE && len >= 2 ) {
		*evenp++ = *buf++;
		*oddp++ = *buf++;
		len -= 2;
	}
	if (evenp < evenbuf + BUFSIZE && len >= 1 ) {
		*evenp++ = *buf++;
		len -= 1;
	}

	//assert( len == 0 );

	SHA1Update( &ctx->sha0, evenbuf, evenp-evenbuf );
	SHA1Update( &ctx->sha1, evenbuf, evenp-evenbuf );
	SHA1Update( &ctx->sha2, oddbuf, oddp-oddbuf );
	SHA1Update( &ctx->sha3, oddbuf, oddp-oddbuf );
#endif
}

/*
 * Final wrapup
 */
void shaDoubleFinal(SHADoubleContext *ctx, unsigned char digest[40])
{
	byte sha0Final[SHA_HASHBYTES], sha1Final[SHA_HASHBYTES];
	byte sha2Final[SHA_HASHBYTES], sha3Final[SHA_HASHBYTES];
	byte sha4Final[SHA_HASHBYTES], sha5Final[SHA_HASHBYTES];
	byte sha6Final[SHA_HASHBYTES], sha7Final[SHA_HASHBYTES];
	SHA1_CTX		sha4, sha5, sha6, sha7;
	byte			zeros[7];
	int				i;

	/* Pick up output from the four hash functions */
	SHA1Final( sha0Final, &ctx->sha0 );
	SHA1Final( sha1Final, &ctx->sha1 );
	SHA1Final( sha2Final, &ctx->sha2 );
	SHA1Final( sha3Final, &ctx->sha3 );

	/* XOR sha2 into sha0, and sha3 into sha1 */
	for (i=0; i<SHA_HASHBYTES; ++i) {
		sha0Final[i] ^= sha2Final[i];
		sha1Final[i] ^= sha3Final[i];
	}

	/*
	 * Do a butterfly mix on the output from this pass.  The amount of data
	 * to be hashed is only 40 bytes, passed 20 bytes at a time through four
	 * hashes, so the costs of doing this are small.
	 */
	SHA1Init (&sha4);
	SHA1Init (&sha5);
	SHA1Init (&sha6);
	SHA1Init (&sha7);

	/*
	 * Butterfly consists of four keyed hashes, which we key by prepending
	 * different numbers of bytes of zeros, as we did above.
	 */
	memset( &zeros, 0, sizeof(zeros) );
	SHA1Update( &sha4, zeros, 4 );
	SHA1Update( &sha5, zeros, 5 );
	SHA1Update( &sha6, zeros, 6 );
	SHA1Update( &sha7, zeros, 7 );

	/* Perform hashes for second butterfly */
	SHA1Update( &sha4, sha0Final, SHA_HASHBYTES );
	SHA1Update( &sha5, sha0Final, SHA_HASHBYTES );
	SHA1Update( &sha6, sha1Final, SHA_HASHBYTES );
	SHA1Update( &sha7, sha1Final, SHA_HASHBYTES );
	SHA1Final( sha4Final, &sha4 );
	SHA1Final( sha5Final, &sha5 );
	SHA1Final( sha6Final, &sha6 );
	SHA1Final( sha7Final, &sha7 );

	/* Left half of output is xor of sha4 and sha6 */
	for (i=0; i<SHA_HASHBYTES; ++i)
		sha4Final[i] ^= sha6Final[i];

	/* Right half of output is xor of sha5 and sha7 */
	for (i=0; i<SHA_HASHBYTES; ++i)
		sha5Final[i] ^= sha7Final[i];

	/* Assemble these two halves to produce the final output */
	memcpy(digest, sha4Final, SHA_HASHBYTES);
	memcpy(digest + SHA_HASHBYTES, sha5Final, SHA_HASHBYTES);
}
